#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from .linear_spline import LinearSpline, HingeBasisFunction
from .natural_cubic_spline import NaturalCubicSpline, NaturalCubicSplineBasisFunction
from .util import Timer
